# -*- coding: utf-8 -*-

"""
Alfred Workflow UI simulator.
"""

import typing as T
import time
import subprocess
import dataclasses

import readchar
from blessed import Terminal

from .vendor.os_platform import IS_WINDOWS
from . import exc
from . import events
from .constants import SHOW_ITEMS_LIMIT, SCROLL_SPEED
from .item import T_ITEM, Item
from .line_editor import LineEditor
from .dropdown import Dropdown
from .render import UIRender, T_UI_RENDER
from .debug import debugger
from .ui_formatter import UIFormatterMixin
from .ui_process_key_pressed import UIProcessKeyPressedMixin


key_to_name = {
    readchar.key.CTRL_C: "CTRL_C",
    readchar.key.TAB: "TAB",
    readchar.key.CTRL_X: "CTRL_X",
    readchar.key.UP: "UP",
    readchar.key.DOWN: "DOWN",
    readchar.key.CTRL_E: "CTRL_E",
    readchar.key.CTRL_D: "CTRL_D",
    readchar.key.CTRL_R: "CTRL_R",
    readchar.key.CTRL_F: "CTRL_F",
    readchar.key.LEFT: "LEFT",
    readchar.key.RIGHT: "RIGHT",
    readchar.key.HOME: "HOME",
    readchar.key.END: "END",
    readchar.key.CTRL_H: "CTRL_H",
    readchar.key.CTRL_L: "CTRL_L",
    readchar.key.CTRL_G: "CTRL_G",
    readchar.key.CTRL_K: "CTRL_K",
    readchar.key.BACKSPACE: "BACKSPACE",
    readchar.key.DELETE: "DELETE",
    readchar.key.ENTER: "ENTER",
    readchar.key.CR: "CR",
    readchar.key.LF: "LF",
    readchar.key.CTRL_A: "CTRL_A",
    readchar.key.CTRL_W: "CTRL_W",
    readchar.key.CTRL_P: "CTRL_P",
}

if IS_WINDOWS:
    OPEN_CMD = "start"
else:
    OPEN_CMD = "open"


@dataclasses.dataclass
class DebugItem(Item):
    def enter_handler(self, ui: "UI"):
        subprocess.run([OPEN_CMD, str(debugger.path_log_txt)])


T_HANDLER = T.Callable[[str, T.Optional["UI"]], T.List[T_ITEM]]


class UI(
    UIFormatterMixin,
    UIProcessKeyPressedMixin,
):
    """
    Zelfred terminal UI implementation.

    :param handler: the handler is the core of a Zelfred App. It's a user-defined
        function that takes the entered query and the UI object as inputs
        and returns a list of items to render.
    :param handler: a callable function that takes a query string as input and
        returns a list of items.
    :param hello_message: sometimes the handler execution for the first time
        run takes long. This message will be shown to user to indicate that
        the handler is still running, and the UI will show the returned items
        once the handler is done.
    :param capture_error: whether you want to capture the error and show it
        in the UI, or you want to raise the error immediately.
    :param process_input_immediately: the handler will be called immediately
        after user input. This is the default behavior.
    :param process_input_after_query_delay: the handler will be called after
        a delay of `query_delay` seconds after the first user input. For example,
        let's say the query delay is 0.3 second. At begin, the user input is empty,
        then user entered "a". If then the user entered "bcd" within 0.3 second,
        the handler will not be called to give you the latest items. The handler
        will be called until the user entered a new key, let's say "e", from
        0.4 second after he entered "a".
    :param query_delay: read above.
    :param show_items_limit: max number of items to show in the dropdown menu.
    :param scroll_speed: scroll speed in the dropdown menu.

    Additional private attributes:

    :param _handler_queue: user may nest session in another session, this LIFO
        queue stores the handler functions of the parent sessions, so that
        when we exit the inner session and go back to the previous session.
    :param _line_editor_input_queue: similar to ``_handler_queue``, but stores
        the parent sessions' user input queries.
    """

    def __init__(
        self,
        handler: T_HANDLER,
        hello_message: T.Optional[str] = "Welcome to interactive UI!",
        capture_error: bool = True,
        process_input_immediately: bool = False,
        process_input_after_query_delay: bool = False,
        query_delay: float = 0.3,
        show_items_limit: int = SHOW_ITEMS_LIMIT,
        scroll_speed: int = SCROLL_SPEED,
        terminal: T.Optional[Terminal] = None,
        render_class: T.Type[T_UI_RENDER] = UIRender,
    ):
        # -------------------- input arguments
        self.handler: T_HANDLER = handler
        self._handler_queue: T.List[T_HANDLER] = list()
        self._line_editor_input_queue: T.List[str] = list()
        self.hello_message: T.Optional[str] = hello_message
        self.capture_error: bool = capture_error
        self.query_delay: float = query_delay

        # -------------------- internal implementation related
        if terminal is None:
            self.terminal = Terminal()
        else:
            self.terminal = terminal
        self.render: T_UI_RENDER = render_class(terminal=self.terminal)
        self.event_generator = events.KeyEventGenerator()
        self._process_input: T.Callable
        flag = sum(
            [
                process_input_immediately,
                process_input_after_query_delay,
            ]
        )
        if flag == 0:
            process_input_immediately = True
        elif flag > 1:
            raise ValueError
        else:  # pragma: no cover
            pass

        if process_input_immediately:
            self._process_input = self._process_input_v1_immediately
        elif process_input_after_query_delay:
            self._process_input = self._process_input_v2_after_query_delay
        else:  # pragma: no cover
            raise NotImplementedError

        self._create_key_processor_mapper()
        # -------------------- items related --------------------
        self.line_editor: LineEditor = LineEditor()
        self.dropdown: Dropdown = Dropdown(
            items=[],
            show_items_limit=show_items_limit,
            scroll_speed=scroll_speed,
        )
        self.n_items_on_screen: int = 0

        # -------------------- controller flags --------------------
        self.need_run_handler: bool = True
        self.need_move_to_end: bool = True
        self.need_clear_items: bool = True
        self.need_clear_query: bool = True
        self.need_print_query: bool = True
        self.need_print_items: bool = True
        self.need_process_input: bool = True

    def _debug_controller_flags(self):
        print(f"self.need_run_handler = {self.need_run_handler}")
        print(f"self.need_move_to_end = {self.need_move_to_end}")
        print(f"self.need_clear_items = {self.need_clear_items}")
        print(f"self.need_clear_query = {self.need_clear_query}")
        print(f"self.need_print_query = {self.need_print_query}")
        print(f"self.need_print_items = {self.need_print_items}")
        print(f"self.need_process_input = {self.need_process_input}")

    def replace_handler(self, handler: T_HANDLER):
        """
        Replace the current handler with a new handler, and store the
        current handler and the current user input query, so that we can
        recover them when we need.
        """
        self._handler_queue.append(self.handler)
        self._line_editor_input_queue.append(self.line_editor.line)
        self.handler = handler

    def _clear_query(self):
        """
        Clear the ``(Query): {user_query}`` line.
        """
        self.render.clear_line_editor()
        return True

    def clear_query(self):
        """
        A wrapper of the ``_clear_query()`` method, ensures that the
        ``need_clear_query`` flag is set to ``True`` at the end regardless of
        whether an exception is raised.
        """
        debugger.log("-------------------- clear_query --------------------")
        try:
            if self.need_clear_query:
                debugger.log("before clear")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                self._clear_query()
                debugger.log("after clear")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
            else:
                debugger.log(f"nothing happen")
        finally:
            self.need_clear_query = True

    def _clear_items(self) -> bool:
        """
        Clear the item dropdown menu.
        """
        if self.render.line_number > 1:
            # time.sleep(1)
            self.render.clear_dropdown()
            # time.sleep(1)
            return True
        else:
            return False

    def clear_items(self):
        """
        A wrapper of the ``_clear_items()`` method, ensures that the
        ``need_clear_query`` flag is set to ``True`` at the end regardless of
        whether an exception is raised.
        """
        debugger.log("-------------------- clear_items --------------------")
        try:
            if self.need_clear_items:
                debugger.log("before clear")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                flag = self._clear_items()
                debugger.log("after clear")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                if flag:
                    debugger.log(f"nothing happen")
            else:
                debugger.log(f"nothing happen")
        finally:
            self.need_clear_items = True

    def _print_query(self):
        """
        Print the ``(Query): {user_query}`` line
        """
        content = self.render.print_line_editor(self.line_editor)
        debugger.log(f"printed: {content!r}")

    def print_query(self):
        """
        A wrapper of the core logic for printing query, ensures that the
        ``need_print_query`` flag is set to ``True`` at the end regardless of
        whether an exception is raised.
        """
        debugger.log("-------------------- print_query --------------------")
        try:
            if self.need_print_query:
                debugger.log("before print")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                self._print_query()
                debugger.log("after print")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
            else:
                debugger.log(f"nothing happen")
        finally:
            self.need_print_query = True

    def _print_items(self):
        """
        Core logic for printing items in the dropdown menu.
        """
        debugger.log("render dropdown")
        n_items_on_screen = self.render.print_dropdown(
            self.dropdown, self.render.terminal.width
        )
        self.n_items_on_screen = n_items_on_screen

    def print_items(self):
        """
        A wrapper of the core logic for printint items, ensures that the
        ``need_print_items`` and ``need_run_handler`` flag is set to ``True``
        at the end regardless of whether an exception is raised.
        """
        debugger.log("-------------------- print_items --------------------")
        try:
            if self.need_print_items:
                debugger.log("before print")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                self._print_items()
                # manually move the cursor to the edge to maximize the room for rendering
                self.render.move_down(1)
                self.render.move_up(1)
                debugger.log("after print")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
            else:
                debugger.log(f"nothing happen")
        finally:
            debugger.log(f"move_cursor_to_line_editor ...")
            debugger.log(f"render.line_number: {self.render.line_number}")
            debugger.log(f"render.n_lines: {self.render.n_lines}")
            # always move cursor back to line editor after printing items
            n_vertical, n_horizontal = self.render.move_cursor_to_line_editor(
                self.line_editor
            )
            debugger.log(f"n_vertical: {n_vertical}")
            debugger.log(f"n_horizontal: {n_horizontal}")
            self.need_print_items = True
            self.need_run_handler = True

    # --------------------------------------------------------------------------
    # handle key pressed
    # --------------------------------------------------------------------------
    def process_key_pressed_input(self, pressed: str):
        """
        Process user keyboard input.

        Zelfred default key bindings:

        - ⬆: tap ``Ctrl + E`` or ``UP`` to move item selection cursor up (previous item).
        - ⏫: tap ``Ctrl + R`` to scroll item selection cursor up.
        - ⬇️: tap ``Ctrl + D`` or ``DOWN`` to move item selection cursor down (next item).
        - ⏬: tap ``Ctrl + F`` to scroll item selection cursor down.
        - ⬅️: tap ``LEFT`` to move query input cursor to the left.
        - ➡️: tap ``RIGHT`` to move query input cursor to the right.
        - ⏪: tap ``Alt + LEFT`` to move query input cursor to the previous word.
        - ⏩: tap ``Alt + Right`` to move query input cursor to the next word.
        - ⏮️: tap ``HOME`` to move query input cursor to the beginning of the line.
        - ⏭️: tap ``END`` to move query input cursor to the end of the line.
        - ◀️: tap ``Ctrl + K`` to delete the previous word.
        - ▶️: tap ``Ctrl + L`` to delete the next word.
        - ↩️: tap ``Ctrl + X`` to clear the query input.
        - ◀️: tap ``BACKSPACE`` to delete query input backward.
        - ▶️: tap ``DELETE`` to delete query input forward.
        - 🔴: tap ``Ctrl + C`` to quit the app.
        - ⤴️: tap ``F1`` to quit the current sub session and go back to the previous session and recover previous user input.

        User defined (customizable) item action:

        - 🚀: tap ``Enter`` to perform custom user action.
        - 🚀: tap ``Ctrl + A`` to perform custom item action.
        - 🚀: tap ``Ctrl + W`` to perform custom item action.
        - 🚀: tap ``Ctrl + U`` to perform custom item action.
        - 🚀: tap ``Ctrl + P`` to perform custom item action.

        User defined UI keybinding:

        - 🚀: tap ``CTRL + T`` to perform custom UI action, default is "do nothing.
        - 🚀: tap ``CTRL + G`` to perform custom UI action, default is "do nothing.
        - 🚀: tap ``CTRL + B`` to perform custom UI action, default is "do nothing.
        - 🚀: tap ``CTRL + N`` to perform custom UI action, default is "do nothing.
        """
        pressed_key_name = key_to_name.get(pressed, pressed)
        debugger.log(f"pressed: {pressed_key_name!r}, key code: {pressed!r}")
        self._process_key_pressed_input(pressed)

    def _process_input_v1_immediately(self):
        event = self.event_generator.next()
        if isinstance(event, events.KeyPressedEvent):
            self.process_key_pressed_input(pressed=event.value)

    def _process_input_v2_after_query_delay(self):
        start_time: T.Optional[float] = None
        while 1:
            event = self.event_generator.next()
            if isinstance(event, events.KeyPressedEvent):
                self.process_key_pressed_input(pressed=event.value)
            self.clear_query()
            self.print_query()
            self.render.move_cursor_to_line_editor(self.line_editor)
            if start_time is None:
                start_time = time.time()
            else:
                event_time = time.time()
                if event_time - start_time > self.query_delay:
                    break

    def process_input(self):
        """
        A wrapper of the core logic for processing input, ensures that the
        ``need_process_input`` flag is set to ``True`` at the end regardless of
        whether an exception is raised.
        """
        debugger.log("-------------------- process_input --------------------")
        try:
            if self.need_process_input:
                self._process_input()
        finally:
            self.need_process_input = True

    def run_handler(self, items: T.Optional[T.List[T_ITEM]] = None):
        if self.need_run_handler:
            debugger.log("need to run handler")
            if items is None:
                debugger.log("run handler")
                items = self.handler(self.line_editor.line, self)
            else:
                debugger.log("explicitly give items, skip handler")
            self.dropdown.update(items)

    def move_to_end(self):
        """
        Move the cursor to the next line of the end of the dropdown menu.
        """
        debugger.log("-------------------- move_to_end --------------------")
        try:
            if self.need_move_to_end:
                debugger.log("need to move to end")
                debugger.log("before move to end")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
                n_down = self.render.move_to_end()
                debugger.log(f"move down {n_down} lines")
                debugger.log("after move to end")
                debugger.log(f"render.line_number: {self.render.line_number}")
                debugger.log(f"render.n_lines: {self.render.n_lines}")
            else:
                debugger.log(f"nothing happen")
        finally:
            self.need_move_to_end = True

    def repaint(self):
        """
        Repaint the UI right after the items is ready. This is useful when you want
        to show a message before running the real handler.
        """
        self.move_to_end()
        self.clear_items()
        self.clear_query()
        self.print_query()
        self.print_items()

    def run_sub_session(
        self,
        handler: T_HANDLER,
        initial_query: str = "",
    ):
        """
        Run a sub session with a new handler. User can tap "F1" to go back to the
        previous session. See :meth:`~zelfred.ui_process_key_pressed.UIProcessKeyPressedMixin.process_f1
        and the ``except exc.JumpOutSessionError:`` part in :meth:`~zelfred.ui.UI.run_session`
        to see how it works.

        :param handler: see :class:`UI`
        :param initial_query: the initial user input query in the sub session.
            it is the start-up text in the user input line editor.
        """
        # replace the current handler with the new handler
        self.replace_handler(handler)
        # update the dropdown items and the line editor content
        self.run_handler()
        self.line_editor.clear_line()
        self.line_editor.enter_text(initial_query)
        # re-paint the UI
        self.repaint()
        # enter the main event loop of the sub query session
        self.run_session(_do_init=False)

    def print_hello_items(self):
        items = [
            Item(
                uid="uid-hello-item",
                title="Welcome to the interactive UI",
            )
        ]
        self.dropdown.update(items)
        self.print_items()

    def print_debug_items(self, e: Exception):
        # if dropdown has items, then there must be one selected item
        # include the selected item information in the error message
        if self.dropdown.items:
            selected_item = self.dropdown.selected_item
            title = selected_item.title
            # handle the error in debug loop special case
            if title.startswith("Error on item: "):
                title = title[len("Error on item: ") :]
            processed_title = self.render.process_title(
                title,
                self.render.terminal.width - 15,
            )
            items = [
                DebugItem(
                    uid="uid",
                    title=f"Error on item: {processed_title}",
                    subtitle=f"{e!r}",
                )
            ]
        else:
            items = [
                DebugItem(
                    uid="uid",
                    title=f"Error on item: NA",
                    subtitle=f"{e!r}",
                )
            ]
        self.dropdown.update(items)
        self.print_items()

    def initialize_loop(self):
        """
        Process the first loop of a session. This loop starts with a hello message,
        and then run the handler, then render the UI.
        """
        debugger.log("=== initialize loop start ===")
        self.print_query()  # show initial query, mostly it is empty
        self.print_hello_items()  # show hello items
        self.run_handler()  # run handler using initial query
        self.move_to_end()  # move cursor to the end
        self.clear_items()  # clear items
        self.clear_query()  # clear query
        self.print_query()  # print query
        self.print_items()  # print items
        debugger.log("=== initialize loop end ===")

    def main_loop(self, _ith: int = 0):
        """
        This is the main loop of the UI. It starts with waiting for user input,
        and the run the handler, then render the UI.
        """
        while True:
            _ith += 1
            debugger.log(f"=== {_ith}th main loop start ===")
            self.process_input()
            # self._debug_controller_flags()
            self.run_handler()
            self.move_to_end()
            self.clear_items()
            self.clear_query()
            self.print_query()
            self.print_items()
            debugger.log(f"=== {_ith}th main loop end ===")

    def jump_out_session_loop(self):
        """
        This loop is executed after a :class:`~zelfred.exc.JumpOutSessionError` is raised.

        It pops up the current session from the queue and use the handler
        of previous session to run handler, and then render the UI.
        """
        if self._handler_queue:
            self.handler = self._handler_queue.pop()
        if self._line_editor_input_queue:
            self.line_editor.clear_line()
            self.line_editor.enter_text(self._line_editor_input_queue.pop())

        self.run_handler()
        self.move_to_end()
        self.clear_items()
        self.clear_query()
        self.print_query()
        self.print_items()

    def debug_loop(self, e: Exception):
        """
        Display error message in the UI and wait for new user input.
        New user input may fix the problem or trigger another error.
        """
        debugger.log("=== debug loop start ===")
        self.move_to_end()
        self.clear_items()
        self.clear_query()
        self.print_query()
        self.print_debug_items(e)
        debugger.log("=== debug loop end ===")

    def run_session(self, _do_init: bool = True):
        """
        Run a "session" in the UI. A "session" has a main loop, and it's handler
        to process user input query.

        Read :ref:`ui-event-loop`
        (or `this link <https://zelfred.readthedocs.io/en/latest/03-UI-Event-Loop/index.html>`_)
        for more information.


        ``except exc.JumpOutSessionError:``

        This error will be raised when user
        :meth:`press F1 <zelfred.ui_process_key_pressed.JumpOutSessionError.process_f1>`
        If the current session is a sub session, then it rolls backs the handler
        and line editor input to the previous state. If the current session is
        the main session, then it will do nothing.

        ``except exc.KeyboardInterrupt:``

        This error will be raised when user press ``Ctrl-C``. It will move the
        cursor to the end and prepare to exit the UI.
        """
        try:
            if _do_init:
                self.initialize_loop()
            self.main_loop()
        except exc.EndOfInputError as e:
            return e.selection
        except exc.JumpOutSessionError:
            self.jump_out_session_loop()
            return self.run_session(_do_init=False)
        except KeyboardInterrupt as e:
            self.move_to_end()
            raise e
        except Exception as e:
            if self.capture_error:
                self.debug_loop(e)
                return self.run_session(_do_init=False)
            else:
                raise e

    def run(self):
        """
        Run the UI.
        """
        try:
            self.run_session()
        except KeyboardInterrupt:
            print("🔴 keyboard interrupt, exit.", end="")
        finally:
            print("")


T_UI = T.TypeVar("T_UI", bound=UI)



# -*- coding: utf-8 -*-

"""
You can use one of the following hotkey to do anything using the selected item.

- :meth:`Enter <zelfred.item.Item.enter_handler>`
- :meth:`Ctrl + A <zelfred.item.Item.ctrl_a_handler>`
- :meth:`Ctrl + W <zelfred.item.Item.ctrl_w_handler>`
- :meth:`Ctrl + P <zelfred.item.Item.ctrl_p_handler>`

This module provides some common actions that you can use directly.
"""

import subprocess
from pathlib import Path

try:  # pragma: no cover
    from mac_notifications import client as mac_notification_client

    has_mac_notifications = True
except ImportError:  # pragma: no cover
    has_mac_notifications = False

try:  # pragma: no cover
    import pyperclip

    has_pyperclip = True
except ImportError:  # pragma: no cover
    has_pyperclip = False

from .vendor.os_platform import IS_WINDOWS


def open_url(url: str):  # pragma: no cover
    """
    Open a URL in the default browser.
    """
    if IS_WINDOWS:
        subprocess.run(["start", url], shell=True)
    else:
        subprocess.run(["open", url])


def open_url_or_print(url: str):  # pragma: no cover
    """
    Open a URL in the default browser. If it cannot open the browser, for example,
    if it is in an SSH remote shell, it will print the URL to the terminal.
    """
    try:
        open_url(url)
    except FileNotFoundError as e:
        if "start" in str(e) or "open" in str(e):
            print(
                f"❗ Your system doesn't support open url in browser to clipboard, "
                f"we print it here so you can copy manually."
            )
            print(url)
        else:
            raise e


def open_file(path: Path):  # pragma: no cover
    """
    Open a file in the default application.
    """
    if IS_WINDOWS:
        subprocess.run([str(path)], shell=True)
    else:
        subprocess.run(["open", str(path)])


def open_file_or_print(path: Path):  # pragma: no cover
    """
    Open a file in the default application. If your system doesn't support
    open file, for example, if it is in an SSH remote shell, it will print the path
    to the terminal.
    """
    try:
        open_file(path)
    except FileNotFoundError as e:
        if str(path) in str(e) or "open" in str(e):
            print(
                f"❗ Your system doesn't support open file, "
                f"we print the file path here so you can copy manually."
            )
            print(path)
        else:
            raise e


def copy_text(text: str):  # pragma: no cover
    """
    Copy text to clipboard.

    If your system doesn't support copy to clipboard, for example, if it is
    in an SSH remote shell, it will raise ``pyperclip.PyperclipException``
    """
    if has_pyperclip is False:
        raise ImportError(
            "You need to do 'pip install pyperclip' first to copy text to clipboard."
        )
    pyperclip.copy(text)


def copy_or_print(text: str):  # pragma: no cover
    """
    Copy text to clipboard. If your system doesn't support copy to clipboard,
    for example, if it is in an SSH remote shell, it will print the text
    to the terminal.
    """
    try:
        copy_text(text)
    except pyperclip.PyperclipException:
        print(
            f"❗ Your system doesn't support copy to clipboard, "
            f"we print it here so you can copy manually."
        )
        print(text)


def send_mac_notification(
    title: str,
    subtitle: str,
):  # pragma: no cover
    """
    Send a MAC notification.

    This feature is based on the
    `macos-notifications <https://github.com/Jorricks/macos-notifications>`_
    Python library.

    However, this is not working on MacOS > 11.0, because of the API this library
     is using is deprecated. See this discussion for more details
     https://github.com/Jorricks/macos-notifications/issues/8.
    """
    if has_mac_notifications is False:
        raise ImportError(
            "You need to do 'pip install mac_notifications' first to send notification."
        )

    mac_notification_client.create_notification(
        title=title,
        subtitle=subtitle,
    )



# -*- coding: utf-8 -*-

from sys import platform

from readchar import key

UP = key.UP  # select the previous item in dropdown menu.
DOWN = key.DOWN  # select the next item in dropdown menu.
LEFT = key.LEFT  # move cursor
RIGHT = key.RIGHT  # move cursor

TAB = key.TAB  # auto complete
HOME = key.HOME  # move cursor to the beginning of the line
END = key.END  # move cursor to the end of the line
BACKSPACE = key.BACKSPACE  # delete previous character
DELETE = key.DELETE  # delete next character
ENTER = key.ENTER  # user action
CR = key.CR
LF = key.LF

CTRL_W = key.CTRL_W  # user action
CTRL_E = key.CTRL_E  # select the previous item in dropdown menu.
CTRL_R = key.CTRL_R  # scroll up
CTRL_T = key.CTRL_T  # user custom shortcut
CTRL_U = key.CTRL_U  # user action
CTRL_P = key.CTRL_P  # user action
CTRL_A = key.CTRL_A  # user action
CTRL_D = key.CTRL_D  # select the next item in dropdown menu.
CTRL_F = key.CTRL_F  # scroll down
CTRL_G = key.CTRL_G  # user custom shortcut
CTRL_H = key.CTRL_H  # DON'T USE THIS, IT DOESN'T WORK ON WINDOWS
CTRL_K = key.CTRL_K  # delete previous word
CTRL_L = key.CTRL_L  # delete next word
CTRL_X = key.CTRL_X  # clear the user input.
CTRL_C = key.CTRL_C  # keyboard interrupt
CTRL_B = key.CTRL_B  # user custom shortcut
CTRL_N = key.CTRL_N  # user custom shortcut


if platform.startswith(("linux", "darwin", "freebsd")):
    ALT_LEFT = "\x1bb"  # move to previous word
    ALT_RIGHT = "\x1bf"  # move to next word
elif platform in ("win32", "cygwin"):
    ALT_LEFT = "\x00\x9b"  # move to previous word
    ALT_RIGHT = "\x00\x9d"  # move to next word
else:
    raise NotImplementedError(f"The platform {platform} is not supported yet")


# Looks like it doesn't work on windows.
# ALT_Q = "œ"
# ALT_W = "∑"
# ALT_R = "®"
# ALT_T = "†"
# ALT_O = "ø"
# ALT_P = "π"
# ALT_A = "å"
# ALT_S = "ß"
# ALT_D = "∂"
# ALT_F = "ƒ"
# ALT_G = "©"
# ALT_H = "˙"
# ALT_J = "∆"
# ALT_K = "˚"
# ALT_L = "¬"
# ALT_Z = "Ω"
# ALT_X = "≈"
# ALT_C = "ç"
# ALT_V = "√"
# ALT_B = "∫"
# ALT_M = "µ"

F1 = key.F1  # Jump out the sub-session, return to the previous view.
F2 = key.F2
F3 = key.F3
F4 = key.F4
F5 = key.F5
F6 = key.F6
F7 = key.F7
F8 = key.F8
F9 = key.F9
F10 = key.F10
F11 = key.F11
F12 = key.F12

key_code_to_name_mapper = {
    UP: "UP",
    DOWN: "DOWN",
    LEFT: "LEFT",
    RIGHT: "RIGHT",
    TAB: "TAB",
    HOME: "HOME",
    END: "END",
    BACKSPACE: "BACKSPACE",
    DELETE: "DELETE",
    ENTER: "ENTER",
    CR: "CR",
    LF: "LF",
    CTRL_W: "CTRL_W",
    CTRL_E: "CTRL_E",
    CTRL_R: "CTRL_R",
    CTRL_T: "CTRL_T",
    CTRL_U: "CTRL_U",
    CTRL_P: "CTRL_P",
    CTRL_A: "CTRL_A",
    CTRL_D: "CTRL_D",
    CTRL_F: "CTRL_F",
    CTRL_G: "CTRL_G",
    CTRL_H: "CTRL_H",
    CTRL_K: "CTRL_K",
    CTRL_L: "CTRL_L",
    CTRL_X: "CTRL_X",
    CTRL_B: "CTRL_B",
    CTRL_N: "CTRL_N",
    ALT_LEFT: "ALT_LEFT",
    ALT_RIGHT: "ALT_RIGHT",
    # ALT_Q: "ALT_Q",
    # ALT_W: "ALT_W",
    # ALT_R: "ALT_R",
    # ALT_T: "ALT_T",
    # ALT_O: "ALT_O",
    # ALT_P: "ALT_P",
    # ALT_A: "ALT_A",
    # ALT_S: "ALT_S",
    # ALT_D: "ALT_D",
    # ALT_F: "ALT_F",
    # ALT_G: "ALT_G",
    # ALT_H: "ALT_H",
    # ALT_J: "ALT_J",
    # ALT_K: "ALT_K",
    # ALT_L: "ALT_L",
    # ALT_Z: "ALT_Z",
    # ALT_X: "ALT_X",
    # ALT_C: "ALT_C",
    # ALT_V: "ALT_V",
    # ALT_B: "ALT_B",
    # ALT_M: "ALT_M",
    F1: "F1",
    F2: "F2",
    F3: "F3",
    F4: "F4",
    F5: "F5",
    F6: "F6",
    F7: "F7",
    F8: "F8",
    F9: "F9",
    F10: "F10",
    F11: "F11",
    F12: "F12",
}


def get_key_name_by_code(code: str) -> str:
    return key_code_to_name_mapper[code]



# -*- coding: utf-8 -*-

"""
See :class:`UIFormatterMixin`.
"""

import typing as T

if T.TYPE_CHECKING:
    from .ui import UI


class UIFormatterMixin:
    """
    This mixin class implements helper functions to format terminal output.
    """

    def format_shortcut(self: "UI", key: str) -> str:
        """
        Format a keyboard shortcut key. By default, the color is magenta.
        Example:

            :magenta:`Enter` to open url
        """
        return f"{self.terminal.magenta}{key}{self.terminal.normal}"

    @property
    def TAB(self) -> str:
        return self.format_shortcut("Tab")

    @property
    def ENTER(self) -> str:
        return self.format_shortcut("Enter")

    @property
    def CTRL_A(self) -> str:
        return self.format_shortcut("Ctrl A")

    @property
    def CTRL_W(self) -> str:
        return self.format_shortcut("Ctrl W")

    @property
    def CTRL_U(self) -> str:
        return self.format_shortcut("Ctrl U")

    @property
    def CTRL_P(self) -> str:
        return self.format_shortcut("Ctrl P")

    @property
    def F1(self) -> str:
        return self.format_shortcut("F1")

    @property
    def CTRL_T(self) -> str:
        return self.format_shortcut("Ctrl T")

    @property
    def CTRL_G(self) -> str:
        return self.format_shortcut("Ctrl G")

    @property
    def CTRL_B(self) -> str:
        return self.format_shortcut("Ctrl B")

    @property
    def CTRL_N(self) -> str:
        return self.format_shortcut("Ctrl N")

    def format_highlight(self: "UI", text: str) -> str:
        """
        Highlight a text with terminal color. In this project, the color is cyan.
        Example:

            this is a very :cyan:`Important message`!
        """
        return f"{self.terminal.cyan}{text}{self.terminal.normal}"

    def format_key(self: "UI", key: str) -> str:
        """
        Format a key of key-value-pair text with terminal color. In this project,
        the color is cyan. Example:

            tag :cyan:`environment` = :yellow:`production`
        """
        return self.format_highlight(key)

    def format_value(self: "UI", value: str) -> str:
        """
        Format a value of key-value-pair text with terminal color. In this project,
        the color is yellow. Example:

            tag :cyan:`environment` = :yellow:`production`
        """
        return f"{self.terminal.yellow}{value}{self.terminal.normal}"

    def format_key_value(self: "UI", key: str, value: T.Any) -> str:
        """
        Format a key-value pair text with terminal color. In this project, key is
        in cyan and value is in yellow. Example:

            tag :cyan:`environment` = :yellow:`production`
        """
        return f"{self.format_key(key)} = {self.format_value(value)}"



# -*- coding: utf-8 -*-

"""
Zelfred project related Exceptions.
"""

import typing as T


class EndOfInputError(Exception):
    """
    Raises to indicate the UI should exit.
    """

    def __init__(
        self,
        selection: T.Any,
        message: str = "End of input",
        *args,
    ):
        super().__init__(*args)
        self.selection = selection
        self.message = message


class JumpOutSessionError(Exception):
    """
    Raises to indicate that the UI should quit the current session
    (e.g. current handler) and jump out to the previous session (e.g. previous handler).
    """
    pass


JumpOutLoopError = JumpOutSessionError  # this is for backward compatibility


class TerminalTooSmallError(SystemError):
    """
    Raises to indicate that the terminal size is too small to render the UI.
    """
    pass


class NoItemToSelectError(IndexError):
    """
    todo: doc string here
    """
    pass



# -*- coding: utf-8 -*-

"""
See :class:`UIProcessKeyPressedMixin`.
"""

import typing as T

from . import keyboard
from . import exc

if T.TYPE_CHECKING:
    from .item import T_ITEM
    from .ui import UI


class UIProcessKeyPressedMixin:
    """
    This class implements the key pressed event processing logics.

    **Important**

    All methods whose names start with ``process_`` are abstract methods
    that handle specific key press events. They implement the Default behavior:.
    Experienced developers can override them to customize the behavior.
    """

    def cursor_up_and_down(self: "UI"):
        """
        A helper method should be called after the cursor moves up or down.

        - need_run_handler: False, the user input query is not changed, so we
            don't need to run the handler again.
        - need_move_to_end: True, the dropdown menu may change, so we need to
            move the cursor to the end of the line in the next event loop
        - need_clear_items: True, the dropdown menu may change, so we need to
            clear the dropdown menu.
        - need_clear_query: False, the user input query is not changed, so we
            don't need to clear the user input query.
        - need_print_query: False, the user input query is not changed, so we
            don't need to print the user input query.
        - need_print_items: True, the dropdown menu may change, so we need to
            print the dropdown menu.
        - need_process_input: True, we need to process the user input next time.
        """
        self.need_run_handler: bool = False
        self.need_clear_query: bool = False
        self.need_print_query: bool = False

    def wait_next_user_input(self: "UI"):
        """
        Don't repaint
        """
        self.cursor_left_and_right()

    def cursor_left_and_right(self: "UI"):
        """
        A helper method should be called after the cursor moves left and right.

        - need_run_handler: False, the user input query is not changed, so we
            don't need to run the handler again.
        - need_move_to_end: False, the dropdown menu is not changed, so we
            don't need to move the cursor to the end of the line.
        - need_clear_items: False, the dropdown menu is not changed, so we
            don't need to clear the dropdown menu.
        - need_clear_query: False, the user input query is not changed, so we
            don't need to clear the user input query.
        - need_print_query: False, the user input query is not changed, so we
            don't need to print the user input query.
        - need_print_items: False, the dropdown menu is not change, so we
            don't need to print the dropdown menu.
        - need_process_input: True, we need to process the user input next time.
        """
        self.need_run_handler = False
        self.need_move_to_end = False
        self.need_clear_items = False
        self.need_clear_query = False
        self.need_print_query = False
        self.need_print_items = False

    def process_up(self: "UI"):
        """
        Default behavior:

        select the previous item in dropdown menu.
        """
        self.cursor_up_and_down()
        self.dropdown.press_up()

    def process_down(self: "UI"):
        """
        Default behavior:

        select the next item in dropdown menu.
        """
        self.cursor_up_and_down()
        self.dropdown.press_down()

    def process_ctrl_e(self: "UI"):
        """
        Default behavior:

        select the previous item in dropdown menu.
        """
        self.cursor_up_and_down()
        self.dropdown.press_up()

    def process_ctrl_d(self: "UI"):
        """
        Default behavior:

        select the next item in dropdown menu.
        """
        self.cursor_up_and_down()
        self.dropdown.press_down()

    def process_ctrl_r(self: "UI"):
        """
        Default behavior:

        scroll up the dropdown menu (press UP 5 times).
        """
        self.cursor_up_and_down()
        self.dropdown.scroll_up()

    def process_ctrl_f(self: "UI"):
        """
        Default behavior:

        scroll down the dropdown menu (press DOWN 5 times).
        """
        self.cursor_up_and_down()
        self.dropdown.scroll_down()

    def process_left(self: "UI"):
        """
        Default behavior:

        move user input cursor to the previous character.
        """
        self.cursor_left_and_right()
        self.line_editor.press_left()

    def process_right(self: "UI"):
        """
        Default behavior:

        move user input cursor to the next character.
        """
        self.cursor_left_and_right()
        self.line_editor.press_right()

    def process_alt_left(self: "UI"):
        """
        Default behavior:

        Move cursor to the beginning of previous word.
        """
        self.cursor_left_and_right()
        self.line_editor.move_word_backward()

    def process_alt_right(self: "UI"):
        """
        Default behavior:

        Move cursor to the beginning of next word.
        """
        self.cursor_left_and_right()
        self.line_editor.move_word_forward()

    def process_home(self: "UI"):
        """
        Default behavior:

        Move cursor to the beginning of the line.
        """
        self.cursor_left_and_right()
        self.line_editor.press_home()

    def process_end(self: "UI"):
        """
        Default behavior:

        Move cursor to the end of the line.
        """
        self.cursor_left_and_right()
        self.line_editor.press_end()

    def press_backspace(self: "UI"):
        """
        Default behavior:

        Delete character backwards in the line editor. Also move cursor to the left.
        """
        self.line_editor.press_backspace()

    def press_delete(self: "UI"):
        """
        Default behavior:

        Delete character forwards in the line editor. Also, the cursor stays.
        """
        self.line_editor.press_delete()

    def process_ctrl_k(self: "UI"):
        """
        Default behavior:

        Delete the previous word.
        """
        self.line_editor.delete_word_backward()

    def process_ctrl_l(self: "UI"):
        """
        Default behavior:

        Delete the next word.
        """
        self.line_editor.delete_word_forward()

    def process_tab(self: "UI"):
        """
        Default behavior:

        Auto complete.
        """
        self.line_editor.clear_line()
        selected_item = self.dropdown.selected_item
        if selected_item.autocomplete:
            self.line_editor.enter_text(selected_item.autocomplete)

    def process_ctrl_x(self: "UI"):
        """
        Default behavior:

        Clear the user input.
        """
        self.line_editor.clear_line()

    def process_ctrl_c(self: "UI"):
        """
        Default behavior:

        Keyboard interrupt.
        """
        raise KeyboardInterrupt()

    def validate_selected_item(self: "UI") -> T.Optional["T_ITEM"]:
        if self.dropdown.n_items == 0:
            raise exc.EndOfInputError(
                selection="select nothing",
            )
        else:
            self.move_to_end()
            if self.dropdown.items:
                return self.dropdown.selected_item
        return None

    def process_enter(self: "UI"):
        """
        Default behavior:

        User action for enter.
        """
        selected_item = self.validate_selected_item()
        if selected_item is not None:
            selected_item.enter_handler(ui=self)
            selected_item.post_enter_handler(ui=self)

    def process_ctrl_a(self: "UI"):
        """
        Default behavior:

        User action for CTRL A.
        """
        selected_item = self.validate_selected_item()
        if selected_item is not None:
            selected_item.ctrl_a_handler(ui=self)
            selected_item.post_ctrl_a_handler(ui=self)

    def process_ctrl_w(self: "UI"):
        """
        Default behavior:

        User action for CTRL W.
        """
        selected_item = self.validate_selected_item()
        if selected_item is not None:
            selected_item.ctrl_w_handler(ui=self)
            selected_item.post_ctrl_w_handler(ui=self)

    def process_ctrl_u(self: "UI"):
        """
        Default behavior:

        User action for CTRL U.
        """
        selected_item = self.validate_selected_item()
        if selected_item is not None:
            selected_item.ctrl_u_handler(ui=self)
            selected_item.post_ctrl_u_handler(ui=self)

    def process_ctrl_p(self: "UI"):
        """
        Default behavior:

        User action for CTRL P.
        """
        selected_item = self.validate_selected_item()
        if selected_item is not None:
            selected_item.ctrl_p_handler(ui=self)
            selected_item.post_ctrl_p_handler(ui=self)

    def process_f1(self: "UI"):
        """
        Default behavior:

        Jump out the sub-session, return to the previous view.
        """
        raise exc.JumpOutSessionError

    def process_ctrl_t(self: "UI"):
        """
        Default behavior:

        Do nothing. This is reserved for user custom shortcut.
        """
        self.wait_next_user_input()

    def process_ctrl_g(self: "UI"):
        """
        Default behavior:

        Do nothing. This is reserved for user custom shortcut.
        """
        self.wait_next_user_input()

    def process_ctrl_b(self: "UI"):
        """
        Default behavior:

        Do nothing. This is reserved for user custom shortcut.
        """
        self.wait_next_user_input()

    def process_ctrl_n(self: "UI"):
        """
        Default behavior:

        Do nothing. This is reserved for user custom shortcut.
        """
        self.wait_next_user_input()

    def _create_key_processor_mapper(self):
        """
        Create a key processor mapper. So we can map the key to the corresponding
        processor.
        """
        self._key_processor_mapper = {
            # dropdown menu
            keyboard.UP: self.process_up,
            keyboard.DOWN: self.process_down,
            keyboard.CTRL_E: self.process_ctrl_e,
            keyboard.CTRL_D: self.process_ctrl_d,
            keyboard.CTRL_R: self.process_ctrl_r,
            keyboard.CTRL_F: self.process_ctrl_f,
            # line editor - move cursor
            keyboard.LEFT: self.process_left,
            keyboard.RIGHT: self.process_right,
            keyboard.ALT_LEFT: self.process_alt_left,
            keyboard.ALT_RIGHT: self.process_alt_right,
            keyboard.HOME: self.process_home,
            keyboard.END: self.process_end,
            # line editor - change text
            keyboard.BACKSPACE: self.press_backspace,
            keyboard.DELETE: self.press_delete,
            keyboard.CTRL_K: self.process_ctrl_k,
            keyboard.CTRL_L: self.process_ctrl_l,
            keyboard.TAB: self.process_tab,
            keyboard.CTRL_X: self.process_ctrl_x,
            # special
            keyboard.CTRL_C: self.process_ctrl_c,
            keyboard.F1: self.process_f1,
            # User defined (customizable) item action
            keyboard.ENTER: self.process_enter,
            keyboard.CTRL_A: self.process_ctrl_a,
            keyboard.CTRL_W: self.process_ctrl_w,
            keyboard.CTRL_U: self.process_ctrl_u,
            keyboard.CTRL_P: self.process_ctrl_p,
            # User defined UI keybinding:
            keyboard.CTRL_T: self.process_ctrl_t,
            keyboard.CTRL_G: self.process_ctrl_g,
            keyboard.CTRL_B: self.process_ctrl_b,
            keyboard.CTRL_N: self.process_ctrl_n,
        }

    def _process_key_pressed_input(self: "UI", key: str):
        """
        Try to process the key pressed

        :return: a boolean flag indicating whether the key pressed is processed
        """
        if key in self._key_processor_mapper:
            self._key_processor_mapper[key]()
        else:
            self.line_editor.press_key(key)
